
package com.tangosol.examples.jmx;

import java.lang.reflect.Constructor;

import javax.management.Attribute;
import javax.management.MalformedObjectNameException;
import javax.management.MBeanInfo;
import javax.management.MBeanServer;
import javax.management.MBeanServerFactory;
import javax.management.ObjectName;
import javax.management.MBeanAttributeInfo;
import javax.management.Query;
import javax.management.QueryExp;

import java.util.Iterator;
import java.util.Collection;
import java.util.Set;
import java.util.Map;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Comparator;

import com.tangosol.util.Base;


/**
 * Helper for using JMX to manager Coherence.
 *
 * @author mf 2006.05.01
 */
public class JMXHelper extends Base {

    /**
     * Construct a JMXHelper.
     */
    public JMXHelper() {
        m_server = getMBeanServer();
    }


    // ---- MBean accessors for get/set/invoke across all cluster nodes -----

    /**
     * Set the specified attribute on all MBeans referenced by the ObjectNames
     * within the collection.
     *
     * @param collObjectNames  a collection of MBean object names to operate on
     * @param attr             the attribute to set
     */
    public void setAttributeAll(Collection collObjectNames, Attribute attr) {
        Throwable exStored = null;
        MBeanServer server = getServer();
        for (Iterator iter = collObjectNames.iterator(); iter.hasNext(); ) {
            try {
                server.setAttribute((ObjectName)iter.next(), attr);
            } catch (Throwable t) {
                // store exception and continue on other MBeans
                exStored = t;
            }
        }

        // if an exception occured, simply rethrow the last one
        if (exStored != null) {
            throw new RuntimeException("One or more setAttribute calls failed.", 
                                       exStored);
        }
    }

    /**
     * Get the specified attribute from all MBeans referenced by the ObjectNames
     * within the collection.
     *
     * @param collObjectNames  a collection of MBean object names to operate on
     * @param sAttrName        the attribute to set
     *
     * @return a map keyed by ObjectName, containing the results, or exceptions
     */
    public Map getAttributeAll(Collection collObjectNames, String sAttrName) {
        Map mapResult = new HashMap();
        MBeanServer server = getServer();
        for (Iterator iter = collObjectNames.iterator(); iter.hasNext(); ) {
            ObjectName objName = (ObjectName)iter.next();
            Object result = null;
            try {
                result = server.getAttribute(objName, sAttrName);
            } catch (Throwable t) {
                err(t);
            } finally {
                mapResult.put(objName, result);
            }
        }
        return mapResult;
    }

    /**
     * Invoke the specified operation on all MBeans referenced by the
     * ObjectNames within the collection.
     *
     * @param collObjectNames  a collection of MBean object names to operate on
     * @param sOperation       the operation to execute
     * @param aoParams         array containing the operation's parameters
     * @param asSignature      array containing the signature of the operation
     *
     * @return a map keyed by ObjectName, containing the results or exception
     *         from the invocation
     */
    public Map invokeAll(Collection collObjectNames, String sOperation, 
                         Object[] aoParams, String[] asSignature) {
        Map mapResult = new HashMap();
        MBeanServer server = getServer();
        if (aoParams == null) {
            aoParams = new Object[0];
        }
        if (asSignature == null) {
            asSignature = new String[aoParams.length];
            for (int i = 0, c = asSignature.length; i < c; ++i) {
                asSignature[i] = aoParams[i].getClass().getName();
            }
        }
        for (Iterator iter = collObjectNames.iterator(); iter.hasNext(); ) {
            ObjectName objName = (ObjectName)iter.next();
            Object result = null;
            try {
                result = 
                        server.invoke(objName, sOperation, aoParams, asSignature);
            } catch (Throwable t) {
                err(t);
            } finally {
                mapResult.put(objName, result);
            }
        }
        return mapResult;
    }

    /**
     * Find the MBean ObjectName for the cluster node.
     *
     * @param sDomainName  the Coherence domain, or null for the default
     *
     * @return the ObjectNames for the nodes
     */
    public ObjectName getClusterObjectName(String sDomainName) {
        try {
            if (sDomainName == null) {
                sDomainName = DOMAIN_COHERENCE;
            }
            return new ObjectName(sDomainName + CLUSTER_PATTERN);
        } catch (MalformedObjectNameException e) {
            throw Base.ensureRuntimeException(e);
        }
    }

    /**
     * Find all MBean ObjectNames for the cluster's nodes.
     *
     * @param objectnameCluster  the parent Cluster MBean's ObjectName, or null
     *                           for default
     *
     * @return a set of ObjectNames for the nodes
     */
    public Set getNodeSet(ObjectName objectnameCluster) {
        MBeanServer server = getServer();
        try {
            String sPattern = 
                getDomainPattern(objectnameCluster) + NODE_PATTERN;
            return server.queryNames(new ObjectName(sPattern), null);
        } catch (MalformedObjectNameException e) {
            throw Base.ensureRuntimeException(e);
        }
    }

    /**
     * Find all Nodes which are parents of the supplied Coherence MBean ObjectNames.
     *
     * @param setObjectnames  the ObjectNames to use to find the parent node.
     *
     * @return a set of ObjectNames for the nodes
     */
    public Set getParentNodeSet(Set setObjectnames) {
        Set setNodes = new HashSet();
        for (Iterator iter = setObjectnames.iterator(); iter.hasNext(); ) {
            ObjectName objParent = getParentNode((ObjectName)iter.next());
            if (objParent != null) {
                setNodes.add(objParent);
            }
        }
        return setNodes;
    }

    /**
     * Find all Node which is the parent of the supplied Coherence MBean ObjectName.
     *
     * @param objChild  the ObjectName to use to find the parent node.
     *
     * @return an ObjectName for the node, or <tt>null</tt> if none could be found.
     */
    public ObjectName getParentNode(ObjectName objChild) {
        try {
            String sNodeId = objChild.getKeyProperty("nodeId");
            if (sNodeId == null) {
                return null;
            }
            return new ObjectName(objChild.getDomain() + ":type=Node,nodeId=" + 
                                  sNodeId);
        } catch (MalformedObjectNameException e) {
            throw Base.ensureRuntimeException(e);
        }
    }


    /**
     * Find all MBean ObjectNames for a given service.
     *
     * @param objectnameCluster  the parent Cluster MBean's ObjectName, or null
     *                           for default
     * @param sServiceName       the service name, or null for any
     * @param storageEnabled     true to find storage enabled services, false
     *                           for storage disabled services, null for both
     *
     * @return a set of ObjectNames for this service
     */
    public Set getServiceSet(ObjectName objectnameCluster, String sServiceName, 
                             Boolean storageEnabled) {
        MBeanServer server = getServer();
        try {
            String sPattern = 
                getDomainPattern(objectnameCluster) + SERVICE_PATTERN;
            if (sServiceName != null) {
                sPattern += ",name=" + sServiceName;
            }
            QueryExp queryExp = null;
            if (storageEnabled != null) {
                queryExp = 
                        Query.eq(Query.attr("StorageEnabled"), Query.value(storageEnabled.booleanValue()));
            }
            return server.queryNames(new ObjectName(sPattern), queryExp);
        } catch (MalformedObjectNameException e) {
            throw Base.ensureRuntimeException(e);
        }
    }


    /**
     * Find all MBean ObjectNames for a given cache.
     *
     * @param objnameService  the ObjectName of the parent service, or null for
     *                        any service
     * @param sCacheName      the cache name, or null for all caches
     * @param backTier        true to get back tier, false for front tier, or null for both
     *
     * @return a set of ObjectNames for the cache
     */
    public Set getCacheSet(ObjectName objnameService, String sCacheName, 
                           Boolean backTier) {
        MBeanServer server = getServer();
        try {
            String sPattern = getDomainPattern(objnameService) + CACHE_PATTERN;
            if (objnameService != null) {
                sPattern += 
                        ",service=" + objnameService.getKeyProperty("name");
            }
            if (sCacheName != null) {
                sPattern += ",name=" + sCacheName;
            }
            if (backTier != null) {
                sPattern += 
                        ",tier=" + (backTier.booleanValue() ? "back" : "front");
            }
            return server.queryNames(new ObjectName(sPattern), null);
        } catch (MalformedObjectNameException e) {
            throw Base.ensureRuntimeException(e);
        }
    }

    /**
     * Find all MBean ObjectNames for a given StorageManager.
     *
     * @param objnameCache  the ObjectName of the parent cache, or null for
     *                      all caches
     *
     * @return a set of ObjectNames for the StorageManager
     */
    public Set getStorageManagerSet(ObjectName objnameCache) {
        MBeanServer server = getServer();
        try {
            String sPattern = getDomainPattern(objnameCache) + STORAGE_PATTERN;
            if (objnameCache != null) {
                sPattern += 
                        ",service=" + objnameCache.getKeyProperty("service");
                sPattern += ",cache=" + objnameCache.getKeyProperty("name");
            }
            return server.queryNames(new ObjectName(sPattern), null);
        } catch (MalformedObjectNameException e) {
            throw Base.ensureRuntimeException(e);
        }
    }

    /**
     * Filter an ObjectName set based on the supplied query.
     *
     * @param setObjectnames  the set of ObjectNames to filter
     * @param query           the query to use in filtering
     *
     * @return the set of ObjectNames matching the query
     */
    public Set filterSet(Set setObjectnames, QueryExp query) {
        query.setMBeanServer(getMBeanServer());
        Set setResult = new HashSet();

        try {
            for (Iterator iter = setObjectnames.iterator(); iter.hasNext(); ) {
                ObjectName objname = (ObjectName)iter.next();
                if (query.apply(objname)) {
                    setResult.add(objname);
                }
            }
        } catch (Exception e) {
            throw Base.ensureRuntimeException(e);
        }

        return setResult;
    }

    /**
     * Return a reference to the MBeanServer.
     *
     * @return  a reference to the MBeanServer
     */
    public MBeanServer getServer() {
        return m_server;
    }

    /**
     * Return the domain for an ObjectName, or the default Coherence name
     * if the supplied ObjectName is null.
     *
     * @param objectname  the object to retrieve the domain name from
     *
     * @return the domain name
     */
    public String getDomainPattern(ObjectName objectname) {
        return objectname == null ? DOMAIN_COHERENCE : objectname.getDomain();
    }

    // ---- helper methods --------------------------------------------------

    /**
     * Return an MBeanServer for the default JMX management domain.
     *
     * @return the MBeanServer
     */
    protected static MBeanServer getMBeanServer() {
        MBeanServer server = null;
        for (Iterator iter = 
             MBeanServerFactory.findMBeanServer(null).iterator(); 
             iter.hasNext(); ) {
            server = (MBeanServer)iter.next();
            if (DOMAIN_DEFAULT.length() == 0 || 
                server.getDefaultDomain().equals(DOMAIN_DEFAULT)) {
                break;
            }
            server = null;
        }
        if (server == null) {
            server = MBeanServerFactory.createMBeanServer(DOMAIN_DEFAULT);
        }

        return server;
    }


    /**
     * Convert the given object to a value compatible with the specified MBean
     * attribute and return a new Attribute that contains the converted value.
     *
     * @param sAttrName the name of the attribute
     * @param info      the MBeanInfo used to determine the attribute data type
     * @param oValue    the value to convert
     *
     * @return a new Attribute containing the converted value
     */
    public static Attribute createAttribute(String sAttrName, MBeanInfo info, 
                                            Object oValue) {
        try {
            Attribute attr = null;
            Class clazz = null;

            MBeanAttributeInfo[] aAttrInfo = info.getAttributes();
            for (int i = 0; i < aAttrInfo.length; i++) {
                if (aAttrInfo[i].getName().equals(sAttrName)) {
                    clazz = Class.forName(aAttrInfo[i].getType());
                    break;
                }
            }

            if (clazz != null) {
                Constructor constructor = 
                    clazz.getConstructor(new Class[] { oValue.getClass() });
                Object oConverted = 
                    constructor.newInstance(new Object[] { oValue });

                attr = new Attribute(sAttrName, oConverted);
            }
            return attr;
        } catch (Exception e) {
            throw Base.ensureRuntimeException(e);
        }
    }

    public static class ObjectNameComparator implements Comparator {
        /**
         * Construct a new ObjectNameComparator that uses the given key to
         * compare ObjectName objects.
         *
         * @param sKey the key to use in the comparison
         */
        public ObjectNameComparator(String sKey) {
            this.sKey = sKey;
        }

        /**
         * Compare the two ObjectName instances.
         *
         * @param o1 the first ObjectName to compare
         * @param o2 the second ObjectName to compare
         *
         * @return a negative integer, zero, or a positive integer as the first
         *         argument is less than, equal to, or greater than the second.
         */
        public int compare(Object o1, Object o2) {
            ObjectName on1 = (ObjectName)o1;
            ObjectName on2 = (ObjectName)o2;

            String sProp1 = on1.getKeyProperty(sKey);
            String sProp2 = on2.getKeyProperty(sKey);

            Comparable cProp1;
            Comparable cProp2;

            try {
                cProp1 = Integer.valueOf(sProp1);
                cProp2 = Integer.valueOf(sProp2);
            } catch (RuntimeException e) {
                cProp1 = sProp1;
                cProp2 = sProp2;
            }

            int nResult = cProp1.compareTo(cProp2);
            if (nResult == 0) {
                nResult = 
                        on1.getCanonicalName().compareTo(on2.getCanonicalName());
            }

            return nResult;
        }

        /**
         * The key to use in the comparison.
         */
        private String sKey;
    }
    // ---- data members ----------------------------------------------------

    /**
     * The MBeanServer.
     */
    protected MBeanServer m_server;

    // ---- static data members ---------------------------------------------

    /**
     * The default and Coherence JMX management domain names.
     */
    public static final String DOMAIN_DEFAULT = "";
    public static final String DOMAIN_COHERENCE = "Coherence";

    /**
     * Patterns for finding various Coherence MBeans.
     */
    public static final String CLUSTER_PATTERN = ":*,type=Cluster";
    public static final String NODE_PATTERN = ":*,type=Node";
    public static final String SERVICE_PATTERN = ":*,type=Service";
    public static final String CACHE_PATTERN = ":*,type=Cache";
    public static final String STORAGE_PATTERN = ":*,type=StorageManager";
}
