package com.tangosol.examples.trading;

import com.tangosol.examples.data.Trade;

import com.tangosol.net.CacheFactory;
import com.tangosol.net.NamedCache;
import com.tangosol.run.jca.CacheAdapter;

import java.io.IOException;

import java.util.ArrayList;
import java.util.List;

import java.util.Random;

import javax.naming.Context;
import javax.naming.InitialContext;

import javax.naming.NamingException;

import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import javax.transaction.UserTransaction;
/**
 * This an example servlet that demonstrates using coherence with J2CA 
 * intergration and OC4J.
 * @author <a href="mailto:adam.leftik@oracle.com>Adam Leftik</a>
 * @version 1.0
 */
public class TradeServlet extends HttpServlet {
    
    /**
     * Extended from HttpServlet
     * @param req from the servlet engine
     * @param resp from the servlet engine
     * @throws IOException when a i/o error occurs
     */
    public void service(HttpServletRequest req, HttpServletResponse resp) throws IOException {
        String symbol = req.getParameter("symbol");
        String price = req.getParameter("price");
        String lot = req.getParameter("lot");
        
        //by default return ok
        resp.setStatus(HttpServletResponse.SC_OK);
        //check user input
        List<String> errors = validateInput(symbol,price,lot);
        
        if (errors.size() == 0) {
            //construct the domain objec
            Trade trade = new Trade();
            trade.setSymbol(symbol);
            trade.setPrice(Double.parseDouble(price));
            trade.setLot(Integer.parseInt(lot));
            trade.setOrderTime(System.currentTimeMillis());
            //now update the trading system
            performTradeNoTx(trade,errors);
        } 
        
        //send back any error messages or success
        for (String msg : errors) {
           resp.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
           resp.getWriter().println(msg + " ");          
        }
        
        if (errors.size() == 0) {
            resp.getWriter().print("Sucessfully traded");
        }
    }

    /**
     * Get the Coherence Cache Adapter
     * @param ctx the jndi context
     * @return the coherence cache adapter via J2CA
     * @throws NamingException when the J2CA Adapter cannot be found
     */
    private CacheAdapter getCacheAdapter(Context ctx) throws NamingException {
        return new CacheAdapter(ctx, "eis/coherence-tx",CacheAdapter.CONCUR_OPTIMISTIC, CacheAdapter.TRANSACTION_REPEATABLE_GET, -1);   
    }
    
    /**
     * Get a JTA User Transaction
     * @param ctx the jndi context
     * @return a User Transaction
     * @throws NamingException when the user transaction cannot be found
     */
    private UserTransaction getUserTransaction(Context ctx) throws NamingException {
        return (UserTransaction) ctx.lookup("java:comp/UserTransaction");
    }
    
    /**
     * Get jndi context
     * @return the jndi context
     * @throws NamingException when a jndi context cannot be found
     */
    private Context getContext() throws NamingException {
        return new InitialContext();
    }
    
    /**
     * Perforam a trade transaction via J2CA
     * @param trade
     * @param errors
     */
    private void performTradeTx(Trade trade,List<String> errors)  {
    try {
        Context ctx = getContext();
        CacheAdapter cacheFactory = getCacheAdapter(ctx);
        UserTransaction trans = getUserTransaction(ctx);
        //start a jta trans
        trans.begin();
        //grab the cache
        NamedCache cache = cacheFactory.getDistributedCache("demo", Thread.currentThread().getContextClassLoader());
        
        //now update the cache
        Random random = new Random();
        cache.put(random.nextInt(cache.size()-1),trade);
        
        //commit the transaction
        trans.commit();
    
        } catch (Exception e) {
            e.printStackTrace(System.out);
            errors.add(e.getMessage());
        }
        
    }
    
    
    private void performTradeNoTx(Trade trade,List<String> errors)  {
    try {
        NamedCache cache = CacheFactory.getCache("demo", Thread.currentThread().getContextClassLoader());
        
        //now update the cache
        Random random = new Random();
        cache.put(random.nextInt(cache.size()-1),trade);
    
        } catch (Exception e) {
            e.printStackTrace(System.out);
            errors.add(e.getMessage());
        }
        
    }    
    
    /**
     * Validate the query/post strings
     * @param symbol
     * @param price
     * @param lot
     * @return
     */
    private List<String> validateInput(String symbol, String price, String lot) {
        List <String> errors = new ArrayList<String>();
        
        validateSymbol(symbol, errors);
        
        try {
            double d = Double.parseDouble(price);
            if (!(d > 0)) {
               errors.add(" You must enter a positive $");
            }
        } catch (Exception ex) {
           errors.add("You must enter a valid price like 32.00");
        }
        
        try {
            int l  = Integer.parseInt(lot);
        } catch (Exception ex) {
            errors.add("You must enter a lot size of > 0");
        }
        
        return errors;
    }
    
    /**
     * Validate that the symbol exists
     * @param symbol
     * @param errors
     */
    private void validateSymbol(String symbol,List<String> errors) {
        for (String sym: Trade.SYMBOL) {
            if (sym.equals(symbol)) {
                return;
            }
        }
        StringBuilder buf = new StringBuilder("You must enter a valid trade symbol like : ");
        for (String s : Trade.SYMBOL) { 
            buf.append(s  + " ");
        }
        errors.add(buf.toString());
    }    
}
